/* AC discrimination Nets*/

:- import append/3 from basics.
%:- table gen_start_state/2.

gen_trans(P) :-
	initialize,
	gen_start_state(proc(P), Start),
	assert(start_state(Start)),
	index(trans/5,trie),
	gen(proc(P),trans(S,A,C,T)),
	not(res_varact(A)),
	translate(trans(S,A,C,T), trans(S1,A1,C1,T1)),
	get_rule_number(N),
	assert(trans(N,S1,A1,C1,T1)),
	assert_terms(S1,N),
	fail
	; true.

assert_terms(S, C1) :-
        S \= prod(_,_),!,
        assert(rule(S, C1)).
assert_terms(prod(S1,S2), C1) :-
        assert(rule(S1, C1)),
        assert(rule(S2, C1)).


test_assert(':-'(H,B))  :- !, (clause(H,B), ! ; assert(':-'(H,B))).
test_assert(X)          :- clause(X, true), ! ; assert(X).


get_rule_number(N1) :-
        conget(transnumber, N),
        N1 is N+1,
        conset(transnumber, N1).


/* modify trans: given a transition S --A,C--> D, and two states S0 and S1,
   generate a new transition as follows:
   (1) S --A,C--> D unmodified.
   (2) S0 --A,C'--> D if S can unify with S1, and C' is a modification of
	   constraints in C.
   This predicate is used to generate transitions for a process P, given
   transitions for a subprocess P' where P = code(C,P'), P = match(C,P'),
   P = proc(Q) and Q=def=P', or P = choice(P',P''). Note that in all these
   cases, the start states of P and P' differ (and correspond to states
   S0 and S1 in the above explanation) but the actions from P and P' differ
   (possibly) on the constraints imposed by the computations.
*/
	  
/* 
modify_trans(trans(S,A,C1,D), _StartState, _MidState, _Modification,
	     trans(S,A,C2,D)) :-
	check_c(C1, C2).
*/

modify_trans(Trans1, StartState, MidState, Modification, Proc, Trans) :-
	Trans1 = trans(MidState1, A, Cs, NextState1),
	unify_states(MidState1, MidState, NextState1, NextState),
	modify_constraints(Modification, Cs, Cs1),
	check_c(Cs1, Cs2),
	(short_cut_proc(Proc) ->
		(Trans = Trans1
		 ;
		 Trans = trans(StartState, A, Cs2, NextState)
		)
		;
		Trans = trans(StartState, A, Cs2, NextState)
	).

modify_trans(trans(S,A,C1,D), _StartState, MidState, _Proc, _Modification,
	     trans(S,A,C2,D)) :-
	not(unify_states(S, MidState, D, NextState)),
	check_c(C1, C2).

short_cut_proc(code(_,proc(_))) :- !.
short_cut_proc(match(_,proc(_))) :- !.
short_cut_proc(choice(_,proc(_))) :- !.
short_cut_proc(choice(proc(_),_)) :- !.
short_cut_proc(code(_,P)) :- short_cut_proc(P),!.
short_cut_proc(match(_,P)) :- short_cut_proc(P),!.
short_cut_proc(choice(P1,P2)) :- short_cut_proc(P1); short_cut_proc(P2).

modify_constraints(code(C), Cs, [C|Cs]).
modify_constraints(match(X=Y), Cs, Cs1) :-
	(X==Y -> Cs1 = Cs ; Cs1 = [(X=Y)|Cs]).
modify_constraints(none, Cs, Cs).

/** Here begins the compilation rules:
   gen_start_state(P, S): generates the start state S for process P
   gen(P, T): generates a transition T for process P.
**/

%zero
% no transitions for zero
gen_start_state(zero, zero).

% Prefix
/*gen(P1, Trans) :-
	P1 = pref(_A, P),
	gen(P, Trans).*/
gen(P1, Trans) :-
	P1 = pref(A,P),
	gen_start_state(P, MidState),
	gen_start_state(P1, StartState),
	(Trans = trans(StartState, A, [], MidState)
	;
	gen(P,Trans)).
gen_start_state(P1, StartState) :-
	P1 = pref(A,P),
	new_state(P1, StartState).

% Code
gen(P1, Trans) :-
	P1 = code(C,P),
	gen_start_state(P1, StartState),
	gen_start_state(P, MidState),
	gen(P, Trans1),
	modify_trans(Trans1, StartState, MidState, code(C), P1, Trans).
gen_start_state(code(Comp, P), StartState) :-
	new_state(code(Comp, P), StartState).

% Conditional: treated as choice+code
gen(if_then_else(Comp, P1, P2), Trans) :-
	gen(choice(code(Comp, P1), code(not(Comp), P2)), Trans).
gen_start_state(if_then_else(Comp, P1, P2), State) :-
	gen_start_state(choice(code(Comp, P1), code(not(Comp), P2)), State).

% Choice
gen(P, Trans) :-
	P = choice(P1, P2),
	gen_start_state(P, StartState),
	gen_start_state(P1, MidState),
	gen(P1, Trans1),
	modify_trans(Trans1, StartState, MidState, none, P, Trans).
gen(P, Trans) :-
	P = choice(P1, P2),
	gen_start_state(P, StartState),
	gen_start_state(P2, MidState),
	gen(P2, Trans1),
	modify_trans(Trans1, StartState, MidState, none, P, Trans).
gen_start_state(P, State) :-
	P = choice(P1,P2),
	new_state(P, State).

% Res
gen(nu(X,P), Trans) :-
	gen_new_name(X),
	gen(P, Trans).
%	gen(P, trans(S,A,C,T)),
%	Trans = trans(S,A,[gen_new_name(X)|C],T).
gen_start_state(P, State) :-
	P = nu(X,P1),
	gen_new_name(X),
	gen_start_state(P1, State).

restricted(in(X,_), V) :- X == V.
restricted(out(X,_), V) :- X == V.

% Match
gen(P1, Trans) :-
	P1 = match(X=Y,P),
	gen_start_state(P1, StartState),
	gen_start_state(P, MidState),
	gen(P, Trans1),
	modify_trans(Trans1, StartState, MidState, match(X=Y), P1, Trans).
gen_start_state(P1, Start) :-
	P1 = match(X=Y,P),
	new_state(P1, Start).

% Par
gen(par(P1, _P2), Trans) :-
	gen(P1, Trans1),
	make_prod_trans1(Trans1, Trans).
gen(par(_P1, P2), Trans) :-
	gen(P2, Trans2),
	make_prod_trans2(Trans2, Trans).
gen(par(P1, P2), Trans) :-
	gen_prod(P1, P2, Trans).
	%optimize(Trans, ITrans).

:- table gen_prod/3.
gen_prod(P1, P2, Trans) :-
	gen(P1, Trans1),
	Trans1 = trans(S1, A1, C1, T1),
	gen(P2, Trans2),
	Trans2 = trans(S2, A2, C2, T2),
	complement(A1, A2, C3),
	append(C1, C2, C2new),
	append(C2new, C3, C3new),
	check_c(C3new, C4),
	make_prod_state(prod(S1, S2), S),
	make_prod_state(prod(T1, T2), T),
	Trans = trans(S, tau, C4, T).
	%optimize(Trans, ITrans).

gen_start_state(P, S) :- 
	P = par(P1,P2),
	gen_start_state(P1, S1),
	gen_start_state(P2, S2),
	make_prod_state(prod(S1, S2), S).

% Proc
gen(P, Trans) :-
	P = proc(PN),
/*
% Spec
        gen_proc(P, ITrans),
	optimize(Trans, ITrans).
% Spec
*/

% No Spec	
	functor(PN, F, N),
	functor(QN, F, N),
	gen_proc(proc(QN), Trans),
	PN=QN.
	%optimize(Trans, ITrans).
% No Spec

:- table gen_proc/2.
gen_proc(P, Trans) :-
	P = proc(PN),
	def(PN, P1),
	gen(P1, Trans).
	%optimize(Trans, ITrans).

gen_start_state(P, Start) :-
	P = proc(PN),
/*	def(PN, P1),
	gen_start_state(P1, Start).
*/
	functor(PN, F, N),
	functor(QN, F, N),
	def(QN, P1),
	gen_start_state(P1, Start),
	PN = QN.

/**
% No Spec 
	functor(PN, F, N),
	functor(QN, F, N),
	new_state(proc(QN), Start),
	PN=QN.
% No Spec
**/

check_c([], []) :- !.
check_c([X=Y|Rest], C) :- !,
        ((nonvar(X),nonvar(Y)) -> X==Y,
                                check_c(Rest,C)
                             ;
                                (X == Y -> check_c(Rest,C)
                                        ; check_c(Rest, C1),
                                          C = [X=Y|C1]
        )),!.
check_c([Comp|Rest], [Comp|C]) :- 
	check_c(Rest,C).

%com
complement(in(A,X), out(A1,X), V) :-
	(A==A1
	 ->	V = []
	 ;	
		not(atomic(A)), not(atomic(A1)),
		V = [A=A1]
	).

complement(out(A,X), in(A1,X), V) :-
	(A==A1
	 ->	V = []
	 ;	
		not(atomic(A)), not(atomic(A1)),
		V = [A=A1]
	).


gen_new_name(name(_)).

% Generate a new state.  This is used in Prefix rules.
:- table new_state/2.
% Generate a new state, used in process
new_state(Proc, State) :- 
	get_free_vars(Proc, Vars),
	get_new_state(StateName), !,
	State =.. [StateName, Vars].

:- import str_cat/3 from string.
get_new_state(StateName) :-
        state_prefix(SN),
       state_number(N),
        number_chars(N, Nn),
        atom_chars(An, Nn),
        str_cat(SN, An, StateName).
 

state_prefix(state).

get_free_vars(P, V) :-
	get_free_vars(P, [], V).

get_free_vars(zero, V, V).
get_free_vars(pref(in(A,X), P), V1, V2) :-
	get_free_vars(P, V1, U1),
	get_vars(A, U1, U2),
	remove(X, U2, V2).
get_free_vars(pref(out(A,X), P), V1, V2) :-
	get_free_vars(P, V1, U1),
	get_vars(A, U1, U2),
	get_vars(X, U2, V2).
get_free_vars(pref(tau, P), V1, V2) :-
	get_free_vars(P, V1, V2).
get_free_vars(pref(action(_X), P), V1, V2) :- %action has to be ground.
	get_free_vars(P, V1, V2).
get_free_vars(code(C, P), V1, V2) :-
	get_free_vars(P, V1, U),
	get_vars(C, U, V2).
get_free_vars(if_then_else(C, P1, P2), V1, V2) :-
	get_free_vars(P1, V1, U1),
	get_free_vars(P2, U1, U2),
	get_vars(C, U2, V2).
get_free_vars(choice(P1, P2), V1, V2) :-
	get_free_vars(P1, V1, U),
	get_free_vars(P2, U, V2).
get_free_vars(nu(X, P), V1, V2) :-
	get_free_vars(P, V1, U),
	remove(X, U, V2).
get_free_vars(match(C, P), V1, V2) :-
	get_free_vars(P, V1, U),
	get_vars(C, U, V2).
get_free_vars(par(P1, P2), V1, V2) :-
	get_free_vars(P1, V1, U),
	get_free_vars(P2, U, V2).
get_free_vars(proc(P), V1, V2) :-
	get_vars(P, V1, V2).

:- import conget/2, conset/2 from gensym.
 
state_number(N) :-
        conget(statenumber, M),
        N is M+1,
        conset(statenumber, N).
initialize :-
	abolish_all_tables,
	conset(statenumber, 0),
	retractall(trans(_,_,_,_)),
	retractall(start(_)).


list(P) :-
	initialize,
	list_trans(proc(P)).

:- table list_trans/1.

:- import numbervars/1 from num_vars.

list_trans(P) :-
	write('Term: '), writeln(P),
	/*
	writeln('Transitions:'),
	(gen(P,T),
	 numbervars(T),
	 writeln(T),
	 fail; true),
	*/
	writeln('Start States:'),
	(gen_start_state(P,S), numbervars(S), writeln(S), fail; true),
	nl,
	list_next(P).

list_next(proc(P)) :-
	def(P, E),
	list_trans(E).
list_next(pref(_,P)) :-
	list_trans(P).
list_next(choice(P,_)) :-
	list_trans(P).
list_next(choice(_,P)) :-
	list_trans(P).
list_next(nu(X,P)) :-
%	gen_new_name(X),
	list_trans(P).
list_next(match(_,P)) :-
	list_trans(P).
list_next(if_then_else(_,P,_)) :-
	list_trans(P).
list_next(if_then_else(_,_,P)) :-
	list_trans(P).
list_next(par(P,_)) :-
	list_trans(P).
list_next(par(_,P)) :-
	list_trans(P).


% Take a term, and rename every occurrence of X or name(X) to a
% name: Y or name(Y) respectively.
% Collect the list of equalities between the variables that
% will ensure that the resulting term is linear, but still
% equivalent to the original term.


rename(X, DictIn, DictOut, EqIn, EqOut, Y) :-
	(var(X); res_var(X)), !,
	(in(X, DictIn)
	 ->	(var(X) -> Y = Y; Y = name(_)),
		EqOut = [(Y,X)|EqIn]
         ;	X=Y, 
		EqOut = EqIn
	),
	DictOut=[X|DictIn].
rename(T, DictIn, DictOut, EqIn, EqOut, R) :-
	T =.. [F|Args],
	rename_all(Args, DictIn, DictOut, EqIn, EqOut, NewArgs),
	R =.. [F|NewArgs].

rename_all([], D, D, E, E, []).
rename_all([T|Ts], Din, Dout, Ein, Eout, [R|Rs]) :-
	rename(T, Din, Dm, Ein, Em, R),
	rename_all(Ts, Dm, Dout, Em, Eout, Rs).

% Given a term and a dictionary,
%  rename all occurrences of name(_) to a new variable, and
%  collect all name(_) *not* in the given dictionary.
translate(trans(Src1, Act1, Cond1, Dest1),
	  trans(Src2, Act2, Cond2, Dest2)) :-
	rename(Src1, [], _Dm, [], Em, Src2),
%	rename(Act1, Dm, _, Em, En, Act2),
	Act2 = Act1,
	Dest2 = Dest1, 
	generate_equalities(Em, Cond1, Cond2).

generate_equalities([], Cond, Cond).
generate_equalities([(U,V)|Es], CondIn, CondOut) :-
	CondOut = [(U==V)|Cm],
	generate_equalities(Es, CondIn, Cm).

debug_writeln(S) :-
	copy_term(S, T),
	numbervars(T),
	writeln(T).

optimize(P,I) :-
	intern(P, I).
