/* Non-ac */

%unify_states(MidState, MidState, NextState, NextState).

/* Used in unifying two states in compilation time.*/
unify_states(MidState, MidState1, NextState, NextState1) :-
	extract_names(MidState, PNames1),
	extract_names(MidState1, PNames2),
	MidState1 = MidState,
	distinct_names(PNames1),
	distinct_names(PNames2),
	NextState1 = NextState.

extract_names(Term, Names) :-
	extract_names(Term, [], Names).
extract_names(Term, Acc, Names) :-
	(var(Term)
	 ->	Names = Acc
	 ;	(functor(Term, name, 1)
		 ->	add_to_names(Term, Acc, Names)
		 ;	(Term =.. [_|Args],
			 extract_names_list(Args, Acc, Names)
			)
		)
	).
extract_names_list([], Acc, Acc).
extract_names_list([T|Ts], Acc, Names) :-
	extract_names(T, Acc, Inter),
	extract_names_list(Ts, Inter, Names).

:- index add_to_names/3-2.
add_to_names(X, [], [X]).
add_to_names(X, [Y|Ys], R) :-
	(X==Y
	 ->	R = [Y|Ys]
	 ;	R=[Y|R1], add_to_names(X, Ys, R1)
	).
	
:- import numbervars/3 from num_vars.
:- import length/2 from basics.
distinct_names(L) :-
	copy_term(L, C),
	numbervars(C, 0, N),
	length(L, N).		% all variables in L are distinct.


make_prod_trans1(trans(S, A, C, D), trans(prod(S, X), A, C, prod(D, X))).
make_prod_trans2(trans(S, A, C, D), trans(prod(X, S), A, C, prod(X, D))).

make_prod_state(S, S).

reachable_source(PN, StartState, State) :-

	reachable_state(PN, StartState, T),
	unifiable(T, State).

:- table reachable_state/3.
reachable_state(_PN, S, T) :- make_skel(S, T).
reachable_state(PN, S, T) :-
	gen(proc(PN), Trans),
	Trans = trans(Start, _, _, Dest),
	reachable_state(PN, S, U),
	unifiable(Start, U),
	make_skel(Dest, T).

:- import length/2 from basics.
make_skel(S, T) :- var(S), !, T = T. % Aha, new variable, doent matter.
make_skel(prod(S1, S2), prod(T1, T2)) :- !,
	make_skel(S1, T1),
	make_skel(S2, T2).
make_skel(S, T) :-
	% All states are of the form stateN(VarList), so
	S =.. [F, VL],
	length(VL, N),
	length(Skel, N),
	T =.. [F, Skel].
