:- [utils,mucalculus,count].

init(sym) :-
	[comp,sym, transition].
init(nosym) :-
	[comp,nosym, transition_nosym].
init(net) :-
	[comp_net, sym, transition_net].
init(list) :-
	[comp_list1, sym, transition_list].

version(1, handover,system).
version(2, handover2, system).
version(3, newhandover3, system).
version(4, newhandover7, system(in1, in2, out1, out2)).

formula_version(deadlock, deadlock).
formula_version(ndl, inop(in1, out1)).

gen(Version) :-
	version(Version, File, Start),
	consult(File),
	statistics(0), nl,
	gen_trans(Start),
	statistics,
	writeln('===== compilation done =====').

:- dynamic started/0, done/0.
run(Version, Formula) :-
	retractall(started), retractall(done),
	version(Version, _, Start),
	formula_version(Formula, Form),
	statistics(0), nl,
	write('***** Property '), write(Formula), write(' is '),
	((assert(started), mck1(Start, Form),
	  (done -> true; assert(done), write('TRUE')))
	 ; (started, \+ done, write('FALSE'))
	),
	writeln(' *****'), nl,
	statistics,
	writeln('===== model checking done =====').

stats(SystemOption, Version) :-
	writeln('Statistics:'(SystemOption)),
	version(Version, _, Start),
	get_stats(Start, S, T), nl,
	write('Number of States = '), writeln(S),
	write('Number of Transitions = '), writeln(T),
	specific_stats(SystemOption).

?- [count].
:- import get_calls/3, get_returns/2 from tables.

%specific_stats(net) :- !,
%	write_net_stats(2).
%specific_stats(list) :- !,
%	write_net_stats(2).
%specific_stats(crnet) :- !,
%	write_net_stats(3).
specific_stats(_).

write_net_stats(Arity) :-
	functor(Call, matching_rule_new, Arity),
	count(get_calls(Call, _, _), C),
	write('Number of calls to matching rule (new): '), writeln(C),
	count((get_calls(Call, Cs, Rs), get_returns(Cs, Rs)), A),
	write('Number of answers to matching rule (new): '), writeln(A),
	count(trans(_,_,_,_,_), T),
	write('Number of transitions (new): '), writeln(T),
	Avg is A / C,
	write('Average number of rules tried per call to "transition": '), writeln(Avg),
	Save is Avg * 100 / T,
	write('Average savings over non-net ac unification: '), writeln(Save).
	




all(SystemOption, Version, Formula) :-
	init(SystemOption),
	gen(Version),
	run(Version, Formula),
	stats(SystemOption,Version).



gen1 :- statistics(0), gen_trans(proc(system),system), statistics.

gen2 :- statistics(0), gen_trans(proc(system(in1, in2, out1,out2)),system(in1,in2,out1,out2)),statistics.

r1 :- 
        statistics(0),
	mck(system,deadlock), writeln(yes), statistics, !.
r1 :-  writeln(no1), statistics.


r3 :- 
        statistics(0),
        mck(system(in1, in2, out1, out2),deadlock), writeln(yes), statistics, !.
r3 :-  writeln(no), statistics.

r4 :-
        statistics(0),
	mck(system(in1, in2, out1, out2),inop(in1,out1)), writeln(yes), statistics, !.
r4 :-  writeln(no), statistics.


r2 :- %startstate(system,S),
        statistics(0),
        mck(system,inop(in,out)), writeln(yes), statistics, !.
r2 :-  writeln(no), statistics.
