/* AC + AC net*/

unify_states(MidState1, MidState, NextState1, NextState) :-
	acUnify(MidState1, NextState1, MidState, NextState).
				% if Trans1 starts from MidState
				% modify the constraints appropriately
				% and make it start from StartState

/* This is different from no-sym version and the semantics */
make_prod_trans1(Trans, Trans).
make_prod_trans2(Trans, Trans).

make_prod_state(S, T) :-
	normal_form(S, T).


/* Used in unifying two states in compilation time.*/
acUnify(Pattern1, Pattern2, Subject1, Subject2) :-
	/** +Pattern, +Pattern, +Subject, -Subject **/
%	copy_term((Pattern1, Pattern2), (P1, P2)),
	extract_names(Pattern1, PNames1),
	extract_names(Pattern2, PNames2),
	extract_names(Subject1, SNames1),
	ac_unify(Pattern1, Subject1, Rest),
	distinct_names(PNames1),
	distinct_names(PNames2),
	distinct_names(SNames1),
	(Rest = none
	 ->	Subject2 = Pattern2
	 ;	merge(Pattern2, Rest, Subject2)
	).


extract_names(Term, Names) :-
	extract_names(Term, [], Names).
extract_names(Term, Acc, Names) :-
	(var(Term)
	 ->	Names = Acc
	 ;	(functor(Term, name, 1)
		 ->	add_to_names(Term, Acc, Names)
		 ;	(Term =.. [_|Args],
			 extract_names_list(Args, Acc, Names)
			)
		)
	).
extract_names_list([], Acc, Acc).
extract_names_list([T|Ts], Acc, Names) :-
	extract_names(T, Acc, Inter),
	extract_names_list(Ts, Inter, Names).

:- index add_to_names/3-2.
add_to_names(X, [], [X]).
add_to_names(X, [Y|Ys], R) :-
	(X==Y
	 ->	R = [Y|Ys]
	 ;	R=[Y|R1], add_to_names(X, Ys, R1)
	).
	
:- import numbervars/3 from num_vars.
:- import length/2 from basics.
distinct_names(L) :-
	copy_term(L, C),
	numbervars(C, 0, N),
	length(L, N).		% all variables in L are distinct.

/* residue means the part which cannot be unified*/
ac_unify(prod(P1, RP1), prod(S1, RS1), Residue) :-
	P1 = S1,
	ac_unify(RP1, RS1, Residue).
ac_unify(prod(P1, RP1), prod(S1, RS1), Residue) :- !,
	ac_unify1(P1, RS1, Res1),
	ac_unify(RP1, prod(S1, Res1), Residue).

% The length of lhs must be smaller than that of rhs. 
% P1 must not be of the form prod(_,_).
ac_unify(P1, P2, Residue) :-
	ac_unify1(P1, P2, Residue).

% Residue = none means that all terms are unifiable.
ac_unify1(P1, prod(S1, SR1), SR1) :- 
	P1 = S1.
ac_unify1(P1, prod(S1, SR1), Residue) :- !,
	ac_unify1(P1, SR1, Res1),
	(Res1 == none
		-> Residue = S1
		;  Residue = prod(S1, Res1)
	).
ac_unify1(P, P, none).

normal_form(S,S) :-
	S \= prod(_,_),!.

normal_form(S, NS) :-
	term2List(S, Slist),
	mysort(Slist, Slist1),
	list2Term(Slist1, NS).

mysort(List,Sorted):-q_sort(List,[],Sorted).
q_sort([],Acc,Acc):- !.
q_sort([H|T],Acc,Sorted):-
        pivoting(H,T,L1,L2),
        q_sort(L1,Acc,Sorted1),q_sort(L2,[H|Sorted1],Sorted).
 
pivoting(_H,[],[],[]) :- !.
pivoting(H,[X|T],[X|L],G):- H@=<X,!, pivoting(H,T,L,G).
pivoting(H,[X|T],L,[X|G]):- X@=<H,!, pivoting(H,T,L,G).

term2List(S, L) :-
	term2List(S, [], L).

term2List(S, Acc, [S|Acc]) :-
	(var(S);
         S \= prod(_,_)),!.
term2List(prod(S1, S2), Acc, NS) :-
	term2List(S2, Acc, NS1),
	term2List(S1, NS1, NS).

list2Term([S], S) :-!.
list2Term([A|Rest], prod(A, RestPar)) :-
        list2Term(Rest, RestPar).

merge(prod(X1, S1), prod(X2, S2), S) :- !,
	(X1 @=< X2
	 ->	merge(S1, prod(X2, S2), R),
		S = prod(X1, R)
	 ;	merge(prod(X1, S1), S2, R),
		S = prod(X2, R)
	).
merge(prod(X1, S1), X2, S) :- !,
	(X1 @=< X2
	 ->	merge(S1, X2, R),
		S = prod(X1, R)
	 ;	S = prod(X2, prod(X1, S1))
	).
merge(X1, prod(X2, S2), S) :- !,
	(X1 @=< X2
	 ->	S = prod(X1, prod(X2, S2))
	 ;	merge(X1, S2, R),
		S = prod(X2, R)
	).
merge(X1, X2, S) :-
	(X1 @=< X2
	 ->	S = prod(X1, X2)
	 ;	S = prod(X2, X1)
	).


/*unify(P, S) :- 
	(functor(P, prod, 2)
	-> % proc terms will be subject to ac unification.
	   % First, split P into bound part and variable part
	   extract_bound(P, Bp, Vp),
	   % AC-Unify the bound parts with S, and compute the residue
	   ac_unify(Bp, S, Residue),
	   Residue = Vp
	;  % non-proc terms will be syntactically unified.
		P = S
	).

extract_bound(prod(V1, P1), Bp, Vp) :-
	var(V1), !,
	extract_bound(P1, Bp, Vp1),
	(Vp1 == none
		-> Vp = V1
		;  Vp = prod(V1, Vp1)
	).
extract_bound(P, P, none). */


/*reachable_source(PN, StartState, State) :-
	all_reach(PN, StartState, State).

:- table all_reach/3.
all_reach(PN, S, prod(S1, S2)) :- !,
	all_reach(PN, S, S1),
	all_reach(PN, S, S2).
all_reach(PN, S, T) :-
	reachable_states(PN, S, U),
	\+ ( \+ U = T ). % dont bind, just check for unifiability

:- table reachable_states/3.
reachable_states(_PN, S, T) :-
	contains_state(S, T).
reachable_states(PN, S, T) :-
	gen(proc(PN), Trans),
	Trans = trans(Start, _, _, Dest),
	all_reach(PN, S, Start),
	contains_state(Dest, T).

contains_state(prod(S1, S2), T) :- !,
	(contains_state(S1, T)
	 ;contains_state(S2, T)).
contains_state(S, T) :-
	make_skel(S, T).

:- import length/2 from basics.
make_skel(S, T) :-
	% All states are of the form stateN(VarList), so
	S =.. [F, VL],
	length(VL, N),
	length(Skel, N),
	T =.. [F, Skel].
*/

:- table comp_trans/1.
comp_trans(S) :-
        intern(S, I1),
        transition(I1, A, I2),
        intern(T,I2),
        writeln(trans(S,A,T)),
        nl,
        comp_trans(T).

