/* AC discrimination nets with no rescan of subject term */

start((S,V), I) :-
	term2List(S, L),
	intern((L,V), I).

:- dynamic mark/2.
?- index(mark/2, trie).

used(X) :- assert(mark(used, X)).
succeeded(X) :- assert(mark(succeeded, X)).
selected(X) :- assert(mark(selected, X)).

:- table transition/3.
transition(I1, A, I2) :- 
	intern((S1,L1), I1),
	matching_rule_exact(S1,R,SPos,PPos),
	unify_rule(R, SPos, PPos, S1, L1, A, S2, L2),
	intern((S2, L2), I2).

unify_rule(X, SPos, PPos, S1, L1, B, S2, L2) :-
	trans(X, P1, B, C, P2),
	rule_type(X, Type),
	ac_unify(P1, S1, SPos, PPos, P2, Stemp, Type),
	is_applicable(C, B, L1),
	bubble_sort(Stemp, S2),
	record_outputs(B, L1, S2, L2).

ac_unify(P1, S1, SPos, PPos, P2, Stemp, Type) :-
	(Type == symmetric
	 ->	ac_unify_symmetric(P1, S1, SPos, PPos, P2, Stemp)
	 ;	ac_unify_assymmetric(P1, S1, SPos, PPos, P2, Stemp)
	).

ac_unify_assymmetric([P], S, SPos, _PPos, Pout, Sout) :- !,
	% Single position, so P cannot be of the form prod(_,_).
	select(S, 1, SPos, P, Pout, Sout).
ac_unify_assymmetric(P, S, SPos1, PPos1, Pout, Sout) :-
	% Pair of positions; P is of the form prod(P1, P2)
	sel(P, PPos1, P1, P2),
	select_all(S, 1, [SPos1, _SPos2], [P1,P2], Pout, Sout).

ac_unify_symmetric([P], S, SPos, _PPos, [Pout], Sout) :- !,
	symmetric_select(S, 1, SPos, P, Pout, Sout).
ac_unify_symmetric(P, S, SPos1, PPos1, Pout, Sout) :-
	% Pair of positions; P is of the form prod(P1, P2)
	sel(P, PPos1, P1, P2),
	symmetric_select_all(S, 1, SPos1, [P1,P2], Pout, Sout).

:- index sel/4-2.
sel([P1, P2], 1, P1, P2).
sel([P1, P2], 2, P2, P1).


:- import append/3 from basics.
select([X|Xs], S, I, Y, Add, Out) :- 
	(S == I
	 ->	X=Y, append(Add, Xs, Out)
	 ;	S1 is S + 1, Out=[X|Rs], select(Xs, S1, I, Y, Add,Rs)
	).

symmetric_select([X|Xs], S, I, Y, Repl, Out) :- 
	(S == I
	 ->	X=Y, Out=[Repl|Xs]
	 ;	S1 is S + 1, Out=[X|Rs], symmetric_select(Xs, S1, I, Y, Repl,Rs)
	).

:- index select_all/6-3.
select_all(S, _, [], [], Add, Out) :- append(Add, S, Out).
select_all([X|Xs], Sindex, [I|Is], Selections, Add, Out) :- 
	(Sindex = I,
	 	(Selections = [X|SelRest],
		 Sindex1 is Sindex + 1,
		 select_all(Xs, Sindex1, Is, SelRest, Add, Out))
	 ; (nonvar(I) -> Sindex \= I; true),
		(Sindex1 is Sindex + 1,
		 Out = [X|Rs],
		 select_all(Xs, Sindex1, [I|Is], Selections, Add, Rs))
	).

symmetric_select_all([X|Xs], Sindex, I, Selections, Replacements, Out) :-
	(Sindex == I 
	 ->	Selections = [X|SelRest],
		Replacements = [Y|RepRest],
		Out = [Y|OutRest],
		symmetric_select_rest(Xs, SelRest, RepRest, OutRest)
	 ;	Sindex1 is Sindex + 1,
		Out = [X|Rs],
		symmetric_select_all(Xs, Sindex1, I, Selections, Replacements, Rs)).

/** We know there is only one Sels left, so we can
symmetric_select_rest([], [], [], []).
symmetric_select_rest([X|Xs], [X|Sels], [Y|Reps], [Y|Zs]) :-
	symmetric_select_rest(Xs, Sels, Reps, Zs).
replace the above rules with the following fact:
**/
symmetric_select_rest([X|Xs], [X], [Y], [Y|Xs]).
symmetric_select_rest([X|Xs], Sels, Reps, [X|Zs]) :-
	symmetric_select_rest(Xs, Sels, Reps, Zs).


is_applicable(C, B, L1) :-
	check_cond(C),
	(res_varact(B)
	 ->	in_list(B,L1)
	 ;	true
	).

record_outputs(B, L1, S2, L2) :-
	((B = out(_,Y), nonvar(Y),Y = name(_))
	 ->	add_to_list(Y, L1, Ltmp1)
	 ;	Ltmp1 = L1
        ),
	project_vars(Ltmp1, S2, L2).

project_vars(Ltmp1, S2, L2) :-
        (Ltmp1 == []
	 ->	L2 = []
	 ;	get_vars(S2, Ltmp2),
		intersect(Ltmp1, Ltmp2, L2)
        ).

:- import absmember/2 from listutil.

in_list(in(X,_), L1) :- absmember(X, L1).
in_list(out(X,_), L1) :- absmember(X, L1).

add_to_list(Y, L, R) :-
	(absmember(Y, L)
	 ->	R = L
	 ;	R = [Y|L]
	).

intersect([], _L, []).
intersect([X|Xs], Y, Zs) :-
        (absmember(X, Y)
                ->      Zs = [X|Us]
                ;       Zs = Us
        ), intersect(Xs, Y, Us).


union([], Rlist, Rlist).
union([X|Rest], Rlist1, Rlist) :-
        add_to_names(X, Rlist1, Rlistt),
        union(Rest, Rlistt, Rlist).

:- table matching_rule_new/4.
matching_rule_new(I, C, SPos, PPos) :-
	intern((S,_), I),
	matching_rule_exact(S, C, SPos, PPos).

matching_rule_exact(S1, C, SPos, PPos) :- 
	matching_rule(S1, C, 1, SPos, PPos).

matching_rule([S|_], C, StartPos, StartPos, Pos) :-
	rule(S, C, Pos).
matching_rule([_|Ss], C, StartPos, MatchPos, Pos) :-
	NextPos is StartPos + 1,
	matching_rule(Ss, C, NextPos, MatchPos, Pos).

find_rule(T, C, Pos) :-
	rule(T, C, Pos).

:- import length/2 from basics.
:- dynamic rule_statistics/2.
reset_stats :-
	retractall(rule_statistics(_, _)).

collect_stats(L,M) :-
	length(L, N),
	update_count(N,M).

update_count(N,M1) :-
	(rule_statistics(N, M)
        ->  (retract(rule_statistics(N, _)),
	        M1 is M + 1,
		assert(rule_statistics(N, M1)))
        ;       assert(rule_statistics(N, 1)),
		M1 = 1
        ).






