/* AC discrimination nets*/

start(S, I) :-
	intern(S, I).

:- table transition/3.
transition(I1, A, I2) :-
	intern((S1,L1), I1),
	matching_rule(S1,R),
	unify_rule(R, S1, L1, A, I2).

gen_all([X|R], S1, L1, A, I2) :-
    (
      unify_rule(X, S1, L1, A, I2)
      ;
      gen_all(R, S1, L1, A, I2)
    ).	

unify_rule(X, S1, L1, B, I2) :-
      (	trans(X, P1, B, C, P2),
	ac_unify(P1, S1, Rest),
	check_cond(C),
	(\+ res_varact(B);
	 in_list(B,L1)
	),
	(Rest = none
         ->  normal_form(P2, S2)
         ;   normal_form(prod(P2, Rest), S2)
        ),
	(B = out(_,Y) ->
                ((nonvar(Y),Y = name(_)) -> add_to_list(Y, L1, Ltmp1)
                                ;
                                Ltmp1 = L1
                )
                ;
                Ltmp1 = L1
        ),
        (Ltmp1 = [] -> L2 = []
                    ;
                        get_vars(S2, Ltmp2),
                        intersect(Ltmp2, Ltmp1, L2)
        ),
	intern((S2,L2), I2)
      ).

in_list(in(X,Y), L1) :- !, memberchk(X, L1).
in_list(out(X,Y), L1) :- !, memberchk(X, L1).

memberchk(X, [Y|Rest]) :- X == Y,!.
memberchk(X, [Y|Rest]) :- memberchk(X, Rest).

add_to_list(Y, [], [Y]) :- !.
add_to_list(Y, [X|Rest], [X|Rest]) :- X == Y,!.
add_to_list(Y, [X|Rest], [X|Rest1]) :- add_to_list(Y, Rest, Rest1).

:- import absmember/2 from listutil.

intersect([], _L, []).
intersect([X|Xs], Y, Zs) :-
        (absmember(X, Y)
                ->      Zs = [X|Us]
                ;       Zs = Us
        ), intersect(Xs, Y, Us).


/**
get_trans(S, Rlist) :-
        S \= prod(_,_),
        findall(C, clause(rule(S, C),_), Rlist),!.
get_trans(prod(S1,S2), Rlist) :-
        get_trans(S1, Rlist1),
        get_trans(S2, Rlist2),
        union(Rlist2, Rlist1, Rlist).
**/

union([], Rlist, Rlist).
union([X|Rest], Rlist1, Rlist) :-
        add_to_names(X, Rlist1, Rlistt),
        union(Rest, Rlistt, Rlist).

get_trans(S, Rlist) :-
	copy_term(S,S1),
	setof(C, matching_rule(S1, C), Rlist).
%	collect_stats(Rlist,N).

matching_rule(S, C) :-
	S \= prod(_,_),
	rule(S, C).
matching_rule(prod(S1,S2), C) :-
	matching_rule(S1, C);
	matching_rule(S2, C).


:- import length/2 from basics.
:- dynamic rule_statistics/2.
reset_stats :-
	retractall(rule_statistics(_, _)).

collect_stats(L,M) :-
	length(L, N),
	update_count(N,M).

update_count(N,M1) :-
	(rule_statistics(N, M)
        ->  (retract(rule_statistics(N, _)),
	        M1 is M + 1,
		assert(rule_statistics(N, M1)))
        ;       assert(rule_statistics(N, 1)),
		M1 = 1
        ).






