chan(in(X,Y), X, Y).
chan(out(X,Y), X, Y).

res_var(X) :- nonvar(X), X = name(_).

res_varact(in(X,_Y)) :- res_var(X),!.
res_varact(out(X,_Y)) :- res_var(X).


get_vars(Term, Vars) :- get_vars(Term, [], Vars).
get_vars(Term, Acc, Vars) :-
	(var(Term)
	 ->	insert(Term, Acc, Vars)
	 ;(is_list(Term) -> insert(Term,Acc,Vars)
	 ; (Term = name(_)
	    ->	   insert(Term, Acc, Vars)
	    ;	   Term =.. [_|TermVars],
		   get_list_vars(TermVars, Acc, Vars)
	   )
	)
	).

get_list_vars([], Acc, Acc).
get_list_vars([X|Rest], Acc, Vars) :-
	get_vars(X, Acc, XAcc),
	get_list_vars(Rest, XAcc, Vars).

in(X, [Y|Ys]) :-
	(X==Y
	 ->	true
	 ;	in(X, Ys)
	).

:- index insert/3-2.
insert(X, [], [X]).
insert(X, [Y|L], Acc) :-
	(X==Y
	 ->	Acc=[Y|L]
	 ;	Acc = [Y|R],
		insert(X, L, R)
	).
:- index remove/3-2.
remove(_X, [], []).
remove(X, [Y|L], Res) :-
	(X==Y
	 ->	Res = L
	 ;	remove(X, L, R),
		Res = [Y|R]
	).

remove_all([], L, L).
remove_all([X|Xs], L, R) :-
	remove(X, L, T),
	remove_all(Xs, T, R).

unifiable(S, T) :- \+ ( \+ (S = T)).

check_cond([]).
check_cond([Comp|Rest]):-
	check(Comp),
	check_cond(Rest).
check((X==Y)) :- !, X==Y.
check((X=Y)) :- !,
	((res_var(X), res_var(Y))
	 ->	X == Y
	 ;	X = Y
	).
check(Comp) :-	call(Comp).


:- table myget_transition1/1.

myget_transition :-
        start_state(S),
        start((S,[]), I),
        myget_transition1(I).

myget_transition1(I) :-
        transition(I,A,I1),
        intern((T,[]),I),
        intern((T1,[]),I1),
        writeln('transition('),
        writeln(T),
        writeln(','),
        writeln(A),
        writeln(','),
        writeln(T1),
        writeln(')'),
        nl,
        myget_transition1(I1).


print_transitions :-
   start_state(S),
   start((S,[]), I),
   retractall(start(_)),
   assert(start(I)),
   (state(_), fail; true),
   state(I1),
   intern((S1,_), I1),
   nl,
   print_transitions(I1),
   fail; true.

print_transitions(I1) :-
   transition(I1, A, I2),
   intern((S1, _), I1),
   intern((S2, _), I2),
        writeln('transition('),
        writeln(S1),
        writeln(','),
        writeln(A),
        writeln(','),
        writeln(S2),
        writeln(')'),
        nl,
   fail; true.


:- import trie_intern/3, trie_interned/3 from intern.

intern(Term, InternedValue) :-
	InternedValue = interned_value(Leaf, Skel),
	(nonvar(Leaf)
		-> % lookup from the trie
			trie_interned(Term, Leaf, Skel)
		;  % add to trie
			trie_intern(Term, Leaf, Skel)
	). 
/*
intern(Term, ID, Var) :-
        (var(Term)
                ->  trie_interned(Term, ID, Var)
                ;   (var(ID)
                        -> trie_intern(Term, ID, Var)
                        ; trie_interned(Term, NewId, Var), NewId = ID
                )
        ).

*/

get_stats(_Proc, M, N) :-
	start_state(S),
	start((S,[]), I),
	retractall(start(_)),
	assert(start(I)),
	count(state(_), M),
	count(reachable_transitions(_,_,_), N).

:- import reverse/2 from basics.
bubble_sort(L, S) :-
	bubble_sort_list(L, [], R),
	reverse(R, S).
bubble_sort_list([], S, S).
bubble_sort_list([X|Xs], L, R) :-
	insert_element(X, L, R1),
	bubble_sort_list(Xs, R1, R).

:- index insert_element/3-2.
insert_element(X, [], [X]).
insert_element(X, [Y|Ys], Zs) :-
	(X @< Y
	 ->	Zs = [Y|Zs1], insert_element(X, Ys, Zs1)
	 ;	Zs = [X,Y|Ys]
	).
